=begin
Program:  Atlas-SNP2: A SNP callling and evaluation tool for 454 Life Sciences(Roche) FLX/Titanium and Illumina NGS datas
Version:  1.0 (2010-01-15)
Development Enviroment: Ruby-1.9.1-P376 on Linux
Authors: Jin Yu at Fuli Yu's group, Human Genome Sequencing Center at Baylor College of Medicine
=end

require 'getoptlong'
require 'bigdecimal'


class AtlasSNP2

#Logistic regression coefficients
  $logistic_FLX={:intercept=>-5.36, :quality_score=>0.11, :NQS_pass=>1.07, :swap=>-4.88, :relpos=>0.91}
  $logistic_XLR={:intercept=>-3.31, :quality_score=>0.1, :NQS_pass=>0.26, :swap=>-3.5, :relpos=>-0.37}
  $logistic_SLX={:intercept=>-9.088, :quality_score=>0.162, :NQS_pass=>1.645, :swap=>0.0, :relpos=>2.349}

#Sj priori
  $errorPrior_454_less_3=[0.867, 0.867, 0.062, 0.062, 0.063, 0.063, 0.008, 0.008, 0.0003, 0.0003]
  $snpPrior_454_less_3=[0.78, 0.78, 0.122, 0.122, 0.049, 0.049, 0.049, 0.049, 0.00001, 0.00001]

  $errorPrior_454_greater_3=[0.676, 0.676, 0.102, 0.102, 0.074, 0.074, 0.028, 0.028, 0.12, 0.12]
  $snpPrior_454_greater_3=[0.00001, 0.00001, 0.006, 0.006, 0.006, 0.006, 0.041, 0.041, 0.947, 0.947]

  $errorPrior_SLX=[0.991, 0.004, 0.002, 0.001, 0.001, 0.001, 0.000002, 0.000005, 0.0000001, 0.000001]
  $snpPrior_SLX=[0.0000001, 0.0000001, 0.0000001, 0.0000001, 0.0000001, 0.0000001, 0.0000001, 0.0000001, 0.027, 0.973]
    
  $basechange={"A"=>"T","T"=>"A","G"=>"C","C"=>"G","N"=>"N","X"=>"X","*"=>"*","a"=>"t","t"=>"a","g"=>"c","c"=>"g"}
  
def AtlasSNP2.processArguments()

  opts = GetoptLong.new(
    ["--input", "-i", GetoptLong::REQUIRED_ARGUMENT],
    ["--reference", "-r", GetoptLong::REQUIRED_ARGUMENT],
    ["--help","-h", GetoptLong::NO_ARGUMENT],
    ["--output","-o", GetoptLong::REQUIRED_ARGUMENT],
    ["--maxsub", "-m", GetoptLong::OPTIONAL_ARGUMENT],
    ["--maxindel", "-g", GetoptLong::OPTIONAL_ARGUMENT],
    ["--XLR", "-x", GetoptLong::NO_ARGUMENT],
    ["--Solexa", "-s", GetoptLong::NO_ARGUMENT],
    ["--priorerror","-e", GetoptLong::OPTIONAL_ARGUMENT],
    ["--priorerror_454","-l", GetoptLong::OPTIONAL_ARGUMENT]
  )

  optHash = {}
  opts.each do |opt, arg|
    optHash[opt] = arg
  end

  AtlasSNP2.usage() if (optHash.key?("--help"));
  AtlasSNP2.usage() if (optHash.empty?);
  return optHash
end

def AtlasSNP2.usage(msg='')
    unless (msg.empty?)
      puts "\n#{msg}\n"
    end
      puts "  
Program:  Atlas-SNP2: A SNP callling and evaluation tool for 454 Life Sciences(Roche) FLX/Titanium and Illumina NGS data
Version:  1.0 (2010-01-20)
Development Enviroment: Ruby-1.9.1-P376 on Linux

Usage:	Atlas-SNP2.rb -i [in.sam] -r [reference.fa] -o [results]

-i	FILE	SAM format alignment file (Required)
-r	FILE	FASTA format reference sequence file (Required)
-o	STR	name of output result file (Required)

Choosing Platform: (Default is 454 FLX)
-s 		Illumina
-x 		454 Titanium 

Setting up priori and filters:
-e	FLT	Prior(error|c) when variant coverage number is above 2 (Default is 0.1)
-l	FLT	Prior(error|c) when variant coverage number is 1 or 2 for 454 data (Default is 0.9)
-m 	FLT	maximum percentage of substitution bases in the query (Default is 5.0)
-g 	FLT	maximum percentage of insertion and deletion bases in the query (Default is 5.0)

"
      exit(2);
end

def initialize(optHash)
  @optHash=optHash

  if (!@optHash.key?('--input') or !@optHash.key?("--reference") or !@optHash.key?("--output") )
      AtlasSNP2.usage("Option missing!")
      exit(2);
  end
  
  #set input files

  @SAMFile=@optHash["--input"]
  @refFile=@optHash["--reference"]

  #set parameters
  if @optHash.key?("--maxsub")
      @maxsub = @optHash["--maxsub"].to_f
  else
      @maxsub = 5.0
  end

  if @optHash.key?("--maxindel")
      @maxindel = @optHash["--maxindel"].to_f
  else
      @maxindel = 5.0
  end

  #setup priori parameters

  if @optHash.key?("--priorerror")
      @perror_highcov=@optHash["--priorerror"].to_f
      @psnp_highcov=1-@perror_highcov
  else
      @perror_highcov=0.1
      @psnp_highcov=0.9
  end

  if @optHash.key?("--priorerror_454")
      @perror_lowcov=@optHash["--priorerror_454"].to_f
      @psnp_lowcov=1-@perror_lowcov
  else
      @perror_lowcov=0.9
      @psnp_lowcov=0.1
  end

  if @optHash.key?("--XLR")
      @platform="454_Titanium"
      @logistic=$logistic_XLR
      @errPrior0=$errorPrior_454_greater_3
      @snpPrior0=$snpPrior_454_greater_3
  elsif @optHash.key?("--Solexa")
      @platform="Illumina"
      @logistic=$logistic_SLX
      @errPrior0=$errorPrior_SLX
      @snpPrior0=$snpPrior_SLX
  else
      @platform="454_FLX"
      @logistic=$logistic_FLX
      @errPrior0=$errorPrior_454_greater_3
      @snpPrior0=$snpPrior_454_greater_3
  end

  #set output files
  @outputFile=@optHash["--output"]
  puts "Assuming #{@platform} data"
end

def hashReference()
  ref = ''
  @seq={}
  @snp={}
  @pileup={}
  refReader=File.open(@refFile,"r")
  refReader.each do |line|
    if line =~ /^>(\S+)/
      ref = $1.sub("chr","")
      ref = 'MT' if ref == 'M'
      @seq[ref] = ''
      @snp[ref] = {}
      @pileup[ref] = Hash.new(0)
    else
      @seq[ref] << line.chomp
    end
  end
  refReader.close
 puts "Read reference file completed"
end

def parsing_sam()
    metainfo = 0
    assembly_identifier = []
    num_unmapped = 0
    num_processed = 0
    num_norefence = 0
    sam_Reader=File.open(@SAMFile,"r")
    sam_Reader.each do |line|
      
    if line.match(/^@/)
      metainfo = 1
      if line =~ /AS:(\S+)/
        assembly_identifier << $1 if not assembly_identifier.include?($1)
      end
      next 
    end
    
    read_sam_cols = line.split("\t")
    query = read_sam_cols[0]
    flag = read_sam_cols[1]   
    ref = read_sam_cols[2].sub("chr","")
    offset = read_sam_cols[3].to_i
    cigar = read_sam_cols[5]
    query_seq = read_sam_cols[9]
    query_qual = read_sam_cols[10]
    score = 'NA'
    
    if cigar.count("MIDS") == 0
      num_unmapped += 1
      next
    end
    
    if not @seq.keys.include?(ref)
      num_norefence += 1
      next
    end
    
    #parse optional tags of SAM
    num_fields = read_sam_cols.length
    (11..(num_fields-1)).each do |col_num|
      if read_sam_cols[col_num] =~ /^AS:i:(\S+)$/
        score = $1
      end
    end
    
    query_length = query_seq.length

    #parse bitwise flag to get strand dir, dir == '1' means reversed strand
    if flag.to_i.to_s(2).reverse[4,1] == '1'
      dir = '-'
    else
      dir = '+'
    end


    # parse CIGAR code, find substitutions, calculate qplace, distance_to_3', readenv, tplace, num_snp, num_ins, num_del, cigar_query_length and covered span on reference
    tplace = offset
    qplace = 1
    span = 0
    cigar_query_length = 0
    num_snp = 0
    num_ins = 0
    num_del = 0
    num_clip = 0
    query_snps = Hash.new([])

    cigar.scan(/([\d]+[MISD])/).each do |part|
      part.to_s.match(/([\d]+)([MIDS])/)
      chunk = $1.to_i 
      case $2
      when "M"
        cigar_query_length += chunk    
        span += chunk
        chunk.times do
          refbase = @seq[ref][tplace - 1,1].upcase
          querybase = query_seq[qplace - 1,1].upcase
          
          if refbase != querybase
            allele = query_seq[qplace - 1]
            qual = query_qual[qplace - 1 ].ord - 33

            #calculate distance to 3`
            if dir == '-'
              dist_3 = qplace - 1
            else
              dist_3 = query_length - qplace
            end

            nqs = Calculate_NQSpass(qplace - 1, dist_3, query_qual, query_length)

            #format readEnv
            if qplace >= 7
              readenv = query_seq[(qplace - 7)..(qplace - 2)].to_s.downcase + querybase + query_seq[qplace..(qplace + 5)].to_s.downcase
            elsif qplace == 1
              readenv = querybase + query_seq[qplace..(qplace + 5)].to_s.downcase
            else
              readenv = query_seq[0..(qplace - 2)].to_s.downcase + querybase + query_seq[qplace..(qplace + 5)].to_s.downcase
            end
 
            if dir == '-'
              readenv.reverse!
              readenv.gsub!(/[atcgATCGNX*]/) {|s| $basechange[s[0,1]]}
            end
            
            query_snps[tplace] = [allele, qual, qplace, dist_3, nqs, readenv]
            num_snp += 1
          end
          @pileup[ref][tplace] += 1
          qplace += 1
          tplace += 1
        end
      when "I"
        qplace += chunk
        cigar_query_length += chunk
        num_ins += chunk
      when "D"
        chunk.times do
          @pileup[ref][tplace] += 1
          tplace += 1
        end
        span += chunk
        num_del += chunk
      when "S"
        qplace += chunk
        cigar_query_length += chunk
        num_clip += chunk
      end
    end
    #All substitutions in this query are stored in query_snps[pos][readinfo]
    
    gap = num_ins + num_del
    sub_perc = (num_snp.to_f / (query_length - num_clip) * 10000).round/100.0
    gap_perc = (gap.to_f / (query_length - num_clip) * 10000).round/100.0
    
    if (sub_perc > @maxsub) or (gap_perc > @maxindel)
      next
    end

    if cigar_query_length != query_length
      $stderr.puts "Check CIGAR code failed"
      $stderr.puts line
      next
    end
    readinfo = [query, dir, query_length, score, sub_perc, gap_perc]

    #using all substitutions in this query to calculate whether a substitution is a SWAP, SNP or MNP
    query_snps.each_key do |pos|
      refbase = @seq[ref][pos-1,1].upcase
      curbase = query_snps[pos][0]
      next if (refbase == 'N' or curbase == 'N')
      if query_snps.key?(pos + 1) or query_snps.key?(pos + 2) or query_snps.key?(pos - 1) or query_snps.key?(pos - 2)
        if query_snps.key?(pos + 1) and refbase == query_snps[pos+1][0] and curbase == @seq[ref][pos,1].upcase
         query_snps[pos] << "swap"
        elsif query_snps.key?(pos - 1) and refbase == query_snps[pos-1][0] and curbase == @seq[ref][pos-2,1].upcase 
         query_snps[pos] << "swap"
        elsif query_snps.key?(pos + 2) and refbase == query_snps[pos+2][0] and curbase == @seq[ref][pos+1,1].upcase
         query_snps[pos] << "swap"
        elsif query_snps.key?(pos - 2) and refbase == query_snps[pos-2][0] and curbase == @seq[ref][pos-3,1].upcase
         query_snps[pos] << "swap"
        elsif (query_snps.key?(pos + 1) and query_snps[pos + 1][0] != 'N') or (query_snps.key?(pos - 1) and query_snps[pos - 1][0] != 'N')
         query_snps[pos] << "mnp"
        else
          query_snps[pos] << "snp"
        end
      else
        query_snps[pos] << "snp"
      end
      #complete calculations of all candidate SNPs found in this query, pass them to a global varible @snp[ref][pos]=[[snp1_readinfo],[snp2_readinfo]..]
      @snp[ref][pos] = [] unless @snp[ref].key?(pos)
      @snp[ref][pos] << query_snps[pos] + readinfo
    end
      num_processed += 1
  end
sam_Reader.close

if metainfo == 1 and (not assembly_identifier.empty?)
  puts "Genome assembly identifier: #{assembly_identifier.to_s}"
elsif metainfo == 0
  puts "No header in this SAM file"
end

puts "#{num_unmapped} alignments have no CIGAR code"
puts "#{num_norefence} alignments have no reference"
puts "#{num_processed} alignments are processed"
end

def snp_evaluate()
  outputWriter=File.new(@outputFile,"w")
  outputWriter.print "refName\tcoordinate\trefBase\tvariantBase\toriQual\tvariantReadCov\tnumAlternativeReads\ttotalCoverage\tPr(error)j\tPr(SNP)j\tPr(Sj|error,c)\tPr(Sj|SNP,c)\tPrior(error|c)\tPrior(SNP|c)\tPosterior(SNP|Sj,c)j\trefEnv\thomopolymer\treadsInfo\n"

  @snp.keys.sort.each do |ref|
    @snp[ref].keys.sort.each do |pos|   
      readinfo_output = '' 
      added_qual = 0
      refbase=@seq[ref][pos.to_i - 1, 1]
      
      #format refenv
      if pos.to_i < 7
        refenv = @seq[ref][0, pos.to_i + 6]
      else
        refenv = @seq[ref][pos.to_i - 7, 13]
      end
      
      totalcoverage = @pileup[ref][pos]
      alternativereads=@snp[ref][pos].length
      homonum = homocount(refenv)

      base_num = {}
      @snp[ref][pos].each do |col|   
        base=col[0]
        if base_num[base] == nil
          base_num[base] = {}
          base_num[base][:num] = 1
        else
          base_num[base][:num] += 1
        end
      end
      array = base_num.keys.sort {|a,b| base_num[b][:num] <=> base_num[a][:num]}
      varbase = array[0]
      
      var_cov=0
      rstr=''
      pr_error_j=1
      @snp[ref][pos].each do |read_col|
        env  = read_col[5]
        base = read_col[0]
        qual = read_col[1]
        name = read_col[7]
        score= read_col[10]
        sub  = read_col[11]
        gap  = read_col[12]
        nqs  = read_col[4]
        dist = read_col[3]
        len  = read_col[9]
        dir  = read_col[8]
        swap = read_col[6]
        relpos=dist/len.to_f


        #format readinfo of each substitution
        readinfo_output = "#{base}(#{qual})#{name}(#{dist})(#{score}/#{len})#{dir}#{env}(#{sub}/#{gap})(#{nqs}/#{len})#{swap}"

        if swap =="swap" 
          swap = 1
        elsif swap =="mnp"
          swap = 0.5
        else
          swap = 0
        end
        
        #Only consider queries harbors the same variant base in Bayesian framework
        if base == varbase
          added_qual += qual                                        
          #Logistic regression p-value
          predict = @logistic[:intercept] + @logistic[:quality_score]*qual.to_f + @logistic[:swap]*swap.to_f + @logistic[:NQS_pass]*nqs.to_f + @logistic[:relpos] * relpos.to_f
          pr_snp_i = Math.exp(predict) / (1 + Math.exp(predict))
          pr_error_i=1-pr_snp_i
          pr_error_j*=pr_error_i
          rstr += readinfo_output+"("+pr_error_i.to_s[0..4]+");"
          var_cov+=1
        end
      end

      pr_snp_j=1-BigDecimal(pr_error_j.to_s)
      bin = (BigDecimal(pr_snp_j.to_s) *10).floor

      if bin==10
        bin=9
      end
      
      #set up parameters
      snpPrior = @snpPrior0
      errPrior = @errPrior0
      prior_error_c = @perror_highcov
      prior_snp_c = @psnp_highcov
      
      if (var_cov < 3) and (@platform == "454_FLX" or @platform == "454_Titanium")
        snpPrior = $snpPrior_454_less_3
        errPrior = $errorPrior_454_less_3
        prior_error_c = @perror_lowcov
        prior_snp_c = @psnp_lowcov
      end

      pr_Sj_error_c = errPrior[bin]
      pr_Sj_SNP_c = snpPrior[bin]
      errorPosterior = pr_Sj_error_c * prior_error_c
      snpPosterior = pr_Sj_SNP_c * prior_snp_c
      pr_SNP_Sj_c_j = 1.0 / (1 + errorPosterior/snpPosterior)

      outputWriter.print "chr#{ref}\t#{pos}\t#{refbase}\t#{varbase}\t#{added_qual}\t#{var_cov}\t#{alternativereads}\t#{totalcoverage}\t#{pr_error_j.to_s[0..4]}\t#{pr_snp_j.to_s[0..4]}\t#{pr_Sj_error_c}\t#{pr_Sj_SNP_c}\t#{prior_error_c}\t#{prior_snp_c}\t#{(pr_SNP_Sj_c_j*1000).round/1000.0}\t#{refenv}\t#{homonum}\t#{rstr}\n"
      
    end
  end
  outputWriter.close
end


def Calculate_NQSpass(pos, dist, qual, length)
  p_status = 0
  if length > 10
    if dist > length - 6
      p_status  = 1
    elsif pos >= 5 and length - pos >= 6 and qual[pos].ord >= 53 and qual[pos-1].ord >= 48 and qual[pos-2].ord >= 48 and qual[pos-3].ord >= 48 and qual[pos-4].ord >= 48 and qual[pos-5].ord >= 48 and qual[pos+1].ord >= 48 and qual[pos+2].ord >= 48 and qual[pos+3].ord >= 48 and qual[pos+4].ord >= 48 and qual[pos+5].ord >= 48
      p_status = 1
    end
  end
  return p_status
end

def homocount(str)
  s = str.size
  lasts = ''
  homo = {}
  st = 0
  flag = 0
  0.upto(s-1) do |i|
    cur = str[i,1].upcase
    if cur == lasts
      if flag == 0 # this is the second 
        homo[i-1] = 2
        st = i - 1
      else # this is the third or more
        homo[st] += 1
      end
      flag = 1
    else
      flag = 0
    end
    lasts = cur
  end
  if homo.size > 0
    top = homo.keys.sort {|a,b| homo[b] <=> homo[a]}[0]
    xx = homo[top]
  else
    xx = 1
  end
  return xx
end

#end class definition
end

optHash=AtlasSNP2.processArguments()
start=AtlasSNP2.new(optHash)
start.hashReference()
start.parsing_sam()
start.snp_evaluate()

exit(0);
